from mentat.parsers.file_edit import FileEdit
from mentat.session_input import collect_user_input
from mentat.session_stream import SESSION_STREAM

from .code_file_manager import CODE_FILE_MANAGER
from .conversation import CONVERSATION


async def get_user_feedback_on_edits(
    file_edits: list[FileEdit],
) -> bool:
    stream = SESSION_STREAM.get()
    conversation = CONVERSATION.get()
    code_file_manager = CODE_FILE_MANAGER.get()

    await stream.send(
        "Apply these changes? 'Y/n/i' or provide feedback.",
        color="light_blue",
    )
    user_response_message = await collect_user_input()
    user_response = user_response_message.data

    need_user_request = True
    match user_response.lower():
        case "y" | "":
            edits_to_apply = file_edits
            conversation.add_user_message("User chose to apply all your changes.")
        case "n":
            edits_to_apply = []
            conversation.add_user_message(
                "User chose not to apply any of your changes."
            )
        case "i":
            edits_to_apply = await _user_filter_changes(file_edits)
            conversation.add_user_message(
                "User chose to apply"
                f" {len(edits_to_apply)}/{len(file_edits)} of your suggested"
                " changes."
            )
        case _:
            need_user_request = False
            edits_to_apply = []
            conversation.add_user_message(
                "User chose not to apply any of your changes. User response:"
                f" {user_response}\n\nPlease adjust your previous plan and changes to"
                " reflect this. Respond with a full new set of changes."
            )

    for file_edit in edits_to_apply:
        await file_edit.resolve_conflicts()

    if edits_to_apply:
        await code_file_manager.write_changes_to_files(file_edits=edits_to_apply)
        await stream.send("Changes applied.", color="light_blue")
    else:
        await stream.send("No changes applied.", color="light_blue")

    if need_user_request:
        await stream.send("Can I do anything else for you?", color="light_blue")

    return need_user_request


async def _user_filter_changes(file_edits: list[FileEdit]) -> list[FileEdit]:
    new_edits = list[FileEdit]()
    for file_edit in file_edits:
        if await file_edit.filter_replacements():
            new_edits.append(file_edit)

    return new_edits



import os
from pathlib import Path
from unittest import TestCase

import pytest

from mentat.config_manager import ConfigManager
from mentat.errors import UserError
from mentat.terminal.client import expand_paths


@pytest.mark.asyncio
async def test_path_gitignoring(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    gitignore_path = ".gitignore"
    testing_dir_path = "git_testing_dir"
    os.makedirs(testing_dir_path)

    # create 3 files, 2 ignored in gitignore, 1 not
    ignored_file_path_1 = Path(os.path.join(testing_dir_path, "ignored_file_1.txt"))
    ignored_file_path_2 = Path(os.path.join(testing_dir_path, "ignored_file_2.txt"))
    non_ignored_file_path = Path(os.path.join(testing_dir_path, "non_ignored_file.txt"))

    with open(gitignore_path, "w") as gitignore_file:
        gitignore_file.write("ignored_file_1.txt\nignored_file_2.txt")

    for file_path in [ignored_file_path_1, ignored_file_path_2, non_ignored_file_path]:
        with open(file_path, "w") as file:
            file.write("I am a file")

    # Run CodeFileManager on the git_testing_dir, and also explicitly pass in ignored_file_2.txt
    paths = [testing_dir_path, ignored_file_path_2]
    mock_code_context.settings.paths = paths
    await mock_code_context.refresh(True)

    expected_file_paths = [
        os.path.join(temp_testbed, ignored_file_path_2),
        os.path.join(temp_testbed, non_ignored_file_path),
    ]

    case = TestCase()
    file_paths = [str(file_path.resolve()) for file_path in mock_code_context.files]
    case.assertListEqual(sorted(expected_file_paths), sorted(file_paths))


@pytest.mark.asyncio
async def test_config_glob_exclude(
    mocker, temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Makes sure glob exclude config works
    mock_glob_exclude = mocker.MagicMock()
    mocker.patch.object(ConfigManager, "file_exclude_glob_list", new=mock_glob_exclude)
    mock_glob_exclude.side_effect = [[os.path.join("glob_test", "**", "*.py")]]

    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.py")
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.ts")
    directly_added_glob_excluded_path = Path(
        os.path.join("glob_test", "bagel", "apple", "directly_added_glob_excluded.py")
    )
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    with open(
        directly_added_glob_excluded_path, "w"
    ) as directly_added_glob_excluded_file:
        directly_added_glob_excluded_file.write(
            "Config excludes me but I'm included if added directly"
        )

    mock_code_context.settings.paths = [Path("."), directly_added_glob_excluded_path]
    await mock_code_context.refresh(True)

    file_paths = [str(file_path.resolve()) for file_path in mock_code_context.files]
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_include_path) in file_paths
    assert os.path.join(temp_testbed, directly_added_glob_excluded_path) in file_paths


@pytest.mark.asyncio
async def test_glob_include(temp_testbed, mock_stream, mock_config, mock_code_context):
    # Make sure glob include works
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.py")
    glob_include_path2 = os.path.join("glob_test", "bagel", "apple", "include_me2.py")
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_path), exist_ok=True)
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    os.makedirs(os.path.dirname(glob_include_path2), exist_ok=True)
    with open(glob_include_path2, "w") as glob_include_file:
        glob_include_file.write("I am also included")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am not included")

    file_paths = expand_paths(["**/*.py"])
    mock_code_context.settings.paths = file_paths
    await mock_code_context.refresh(True)

    file_paths = [str(file_path.resolve()) for file_path in mock_code_context.files]
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_include_path) in file_paths
    assert os.path.join(temp_testbed, glob_include_path2) in file_paths


@pytest.mark.asyncio
async def test_cli_glob_exclude(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Make sure cli glob exclude works and overrides regular include
    glob_include_then_exclude_path = os.path.join(
        "glob_test", "bagel", "apple", "include_then_exclude_me.py"
    )
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_then_exclude_path), exist_ok=True)
    with open(glob_include_then_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am included then excluded")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")

    file_paths = expand_paths(["**/*.py"])
    exclude_paths = expand_paths(["**/*.py", "**/*.ts"])
    mock_code_context.settings.paths = file_paths
    mock_code_context.settings.exclude_paths = exclude_paths
    await mock_code_context.refresh(True)

    file_paths = [file_path for file_path in mock_code_context.files]
    assert os.path.join(temp_testbed, glob_include_then_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths


@pytest.mark.asyncio
async def test_text_encoding_checking(
    temp_testbed, mock_stream, mock_config, mock_code_context
):
    # Makes sure we don't include non text encoded files, and we quit if user gives us one
    nontext_path = "iamnottext.py"
    with open(nontext_path, "wb") as f:
        # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
        f.write(bytearray([0x81]))

    mock_code_context.settings.paths = [Path("./")]
    await mock_code_context.refresh(True)
    file_paths = [file_path for file_path in mock_code_context.files]
    assert os.path.join(temp_testbed, nontext_path) not in file_paths

    with pytest.raises(UserError) as e_info:
        nontext_path_requested = "iamalsonottext.py"
        with open(nontext_path_requested, "wb") as f:
            # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
            f.write(bytearray([0x81]))

        paths = [Path(nontext_path_requested)]
        mock_code_context.settings.paths = paths
        await mock_code_context.refresh(True)
    assert e_info.type == UserError



from __future__ import annotations

import os
from pathlib import Path

import attr

from mentat.code_file_manager import CODE_FILE_MANAGER
from mentat.errors import MentatError
from mentat.git_handler import GIT_ROOT
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    FileActionType,
    change_delimiter,
    get_full_change,
)
from mentat.session_input import ask_yes_no
from mentat.session_stream import SESSION_STREAM


# TODO: Add 'owner' to Replacement so that interactive mode can accept/reject multiple replacements at once
@attr.define(order=False)
class Replacement:
    """
    Represents that the lines from starting_line (inclusive) to ending_line (exclusive)
    should be replaced with new_lines
    """

    # Inclusive
    starting_line: int = attr.field()
    # Exclusive
    ending_line: int = attr.field()

    new_lines: list[str] = attr.field()

    def __lt__(self, other: Replacement):
        return self.ending_line < other.ending_line or (
            self.ending_line == other.ending_line
            and self.starting_line < other.ending_line
        )


async def _ask_user_change(
    display_information: DisplayInformation,
    text: str,
) -> bool:
    stream = SESSION_STREAM.get()
    await stream.send(get_full_change(display_information))
    await stream.send(text, color="light_blue")
    return await ask_yes_no(default_yes=True)


@attr.define
class FileEdit:
    """
    Represents that this file_path content should have specified Replacements applied to it.
    Can also represent that this file should be created, deleted, or is being renamed.
    """

    # Should be abs path
    file_path: Path = attr.field()
    replacements: list[Replacement] = attr.field(factory=list)
    is_creation: bool = attr.field(default=False)
    is_deletion: bool = attr.field(default=False)
    # Should be abs path
    rename_file_path: Path | None = attr.field(default=None)

    async def is_valid(self) -> bool:
        stream = SESSION_STREAM.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        git_root = GIT_ROOT.get()

        rel_path = Path(os.path.relpath(self.file_path, git_root))
        if self.is_creation:
            if self.file_path.exists():
                await stream.send(
                    f"File {rel_path} already exists, canceling creation."
                )
                return False
        else:
            if not self.file_path.exists():
                await stream.send(
                    f"File {rel_path} does not exist, canceling all edits to file."
                )
                return False
            elif rel_path not in code_file_manager.file_lines:
                await stream.send(
                    f"File {rel_path} not in context, canceling all edits to file."
                )
                return False

        if self.rename_file_path is not None and self.rename_file_path.exists():
            rel_rename_path = Path(os.path.relpath(self.rename_file_path, git_root))
            await stream.send(
                f"File {rel_path} being renamed to existing file {rel_rename_path},"
                " canceling rename."
            )
            self.rename_file_path = None
        return True

    async def filter_replacements(
        self,
    ) -> bool:
        code_file_manager = CODE_FILE_MANAGER.get()
        git_root = GIT_ROOT.get()

        if self.is_creation:
            display_information = DisplayInformation(
                self.file_path, [], [], [], FileActionType.CreateFile
            )
            if not await _ask_user_change(display_information, "Create this file?"):
                return False
            file_lines = []
        else:
            rel_path = Path(os.path.relpath(self.file_path, git_root))
            file_lines = code_file_manager.file_lines[rel_path]

        if self.is_deletion:
            display_information = DisplayInformation(
                self.file_path, [], [], file_lines, FileActionType.DeleteFile
            )
            if not await _ask_user_change(display_information, "Delete this file?"):
                return False

        if self.rename_file_path is not None:
            display_information = DisplayInformation(
                self.file_path,
                [],
                [],
                [],
                FileActionType.RenameFile,
                new_name=self.rename_file_path,
            )
            if not await _ask_user_change(display_information, "Rename this file?"):
                self.rename_file_path = None

        new_replacements = list[Replacement]()
        for replacement in self.replacements:
            removed_block = file_lines[
                replacement.starting_line : replacement.ending_line
            ]
            display_information = DisplayInformation(
                self.file_path,
                file_lines,
                replacement.new_lines,
                removed_block,
                FileActionType.UpdateFile,
                replacement.starting_line,
                replacement.ending_line,
                self.rename_file_path,
            )
            if await _ask_user_change(display_information, "Keep this change?"):
                new_replacements.append(replacement)
        self.replacements = new_replacements

        return (
            self.is_creation
            or self.is_deletion
            or (self.rename_file_path is not None)
            or len(self.replacements) > 0
        )

    async def _print_resolution(self, first: Replacement, second: Replacement):
        stream = SESSION_STREAM.get()

        await stream.send(
            "Change overlap detected, auto-merged back to back changes:\n"
        )
        await stream.send(self.file_path)
        await stream.send(change_delimiter)
        for line in first.new_lines + second.new_lines:
            await stream.send("+ " + line, color="green")
        await stream.send("")

    async def resolve_conflicts(self):
        self.replacements.sort(reverse=True)
        for index, replacement in enumerate(self.replacements):
            for other in self.replacements[index + 1 :]:
                if (
                    other.ending_line > replacement.starting_line
                    and other.starting_line < replacement.ending_line
                ):
                    # Overlap conflict
                    other.ending_line = replacement.starting_line
                    other.starting_line = min(other.starting_line, other.ending_line)
                    await self._print_resolution(other, replacement)
                elif (
                    other.ending_line == other.starting_line
                    and replacement.ending_line == replacement.starting_line
                    and replacement.starting_line == other.starting_line
                ):
                    # Insertion conflict
                    # This will be a bit wonky if there are more than 2 insertion conflicts on the same line
                    await self._print_resolution(replacement, other)

    def get_updated_file_lines(self, file_lines: list[str]):
        self.replacements.sort(reverse=True)
        earliest_line = None
        for replacement in self.replacements:
            if earliest_line is not None and replacement.ending_line > earliest_line:
                # This should never happen if resolve conflicts is called
                raise MentatError("Error: Line overlap in Replacements")
            if replacement.ending_line > len(file_lines):
                file_lines += [""] * (replacement.ending_line - len(file_lines))
            earliest_line = replacement.starting_line
            file_lines = (
                file_lines[: replacement.starting_line]
                + replacement.new_lines
                + file_lines[replacement.ending_line :]
            )
        return file_lines



import os
import shutil
import stat
import subprocess
import tempfile
from datetime import datetime
from pathlib import Path
from unittest.mock import AsyncMock
from uuid import uuid4

import pytest
import pytest_asyncio

from mentat.code_context import CODE_CONTEXT, CodeContext, CodeContextSettings
from mentat.code_file_manager import CODE_FILE_MANAGER, CodeFileManager
from mentat.config_manager import CONFIG_MANAGER, ConfigManager
from mentat.git_handler import GIT_ROOT
from mentat.session_stream import (
    SESSION_STREAM,
    SessionStream,
    StreamMessage,
    StreamMessageSource,
)
from mentat.streaming_printer import StreamingPrinter

pytest_plugins = ("pytest_reportlog",)


def filter_mark(items, mark, exists):
    new_items = []
    for item in items:
        marker = item.get_closest_marker(mark)
        if bool(marker) == bool(exists):
            new_items.append(item)
    return new_items


def pytest_addoption(parser):
    parser.addoption("--benchmark", action="store_true")
    parser.addoption("--uitest", action="store_true")
    # The following flags are used by benchmark tests
    parser.addoption(
        "--max_exercises",
        action="store",
        default="1",
        help="The maximum number of exercises to run",
    )
    parser.addoption(
        "--max_iterations",
        action="store",
        default="1",
        help="Number of times to rerun mentat with error messages",
    )
    parser.addoption(
        "--language",
        action="store",
        default="python",
        help="Which exercism language to do exercises for",
    )
    parser.addoption(
        "--max_workers",
        action="store",
        default="1",
        help="Number of workers to use for multiprocessing",
    )
    parser.addoption(
        "--refresh_repo",
        action="store_true",
        default=False,
        help="When set local changes will be discarded.",
    )
    parser.addoption(
        "--exercises",
        action="append",
        nargs="*",
        default=[],
        help="Which exercism exercises to run. max_exercises ignored when set.",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "benchmark: run benchmarks that call openai")
    config.addinivalue_line(
        "markers", "uitest: run ui-tests that get evaluated by humans"
    )


def pytest_collection_modifyitems(config, items):
    benchmark = config.getoption("--benchmark")
    uitest = config.getoption("--uitest")
    items[:] = filter_mark(items, "benchmark", benchmark)
    items[:] = filter_mark(items, "uitest", uitest)


@pytest.fixture(scope="function")
def mock_call_llm_api(mocker):
    mock = mocker.patch("mentat.conversation.call_llm_api")

    def set_generator_values(values):
        async def async_generator():
            for value in values:
                yield {"choices": [{"delta": {"content": value}}]}
            yield {"choices": [{"delta": {"content": "\n"}}]}

        mock.return_value = async_generator()

    mock.set_generator_values = set_generator_values

    return mock


@pytest.fixture
def mock_collect_user_input(mocker):
    async_mock = AsyncMock()

    mocker.patch("mentat.code_edit_feedback.collect_user_input", side_effect=async_mock)
    mocker.patch("mentat.session_input.collect_user_input", side_effect=async_mock)
    mocker.patch("mentat.session.collect_user_input", side_effect=async_mock)

    def set_stream_messages(values):
        async_mock.side_effect = [
            StreamMessage(
                id=uuid4(),
                channel="default",
                source=StreamMessageSource.CLIENT,
                data=value,
                extra=None,
                created_at=datetime.utcnow(),
            )
            for value in values
        ]

    async_mock.set_stream_messages = set_stream_messages
    return async_mock


@pytest.fixture(scope="function")
def mock_setup_api_key(mocker):
    mocker.patch("mentat.session.setup_api_key")
    mocker.patch("mentat.conversation.is_model_available")
    return


def add_permissions(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.
    """

    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


# ContextVars need to be set in a synchronous fixture due to pytest not propagating
# async fixture contexts to test contexts.
# https://github.com/pytest-dev/pytest-asyncio/issues/127


# SessionStream
@pytest.fixture
def _mock_stream():
    session_stream = SessionStream()
    token = SESSION_STREAM.set(session_stream)
    yield session_stream
    SESSION_STREAM.reset(token)


@pytest_asyncio.fixture()
async def mock_stream(_mock_stream):
    await _mock_stream.start()
    yield _mock_stream
    await _mock_stream.stop()


# Git root
@pytest.fixture
def _mock_git_root(temp_testbed):
    git_root = Path(temp_testbed)
    token = GIT_ROOT.set(git_root)
    yield git_root
    GIT_ROOT.reset(token)


@pytest_asyncio.fixture()
async def mock_git_root(_mock_git_root):
    yield _mock_git_root


# ConfigManager
@pytest.fixture
def _mock_config():
    config = ConfigManager({}, {})
    token = CONFIG_MANAGER.set(config)
    yield config
    CONFIG_MANAGER.reset(token)


@pytest_asyncio.fixture()
async def mock_config(_mock_config):
    yield _mock_config


# CodeContext
@pytest.fixture
def _mock_code_context(_mock_git_root, _mock_config):
    code_context = CodeContext(settings=CodeContextSettings())
    token = CODE_CONTEXT.set(code_context)
    yield code_context
    CODE_CONTEXT.reset(token)


@pytest_asyncio.fixture()
async def mock_code_context(_mock_code_context):
    await _mock_code_context.refresh(replace_paths=True)
    yield _mock_code_context


# CodeFileManager
@pytest.fixture
def _mock_code_file_manager():
    code_file_manager = CodeFileManager()
    token = CODE_FILE_MANAGER.set(code_file_manager)
    yield code_file_manager
    CODE_FILE_MANAGER.reset(token)


@pytest_asyncio.fixture()
async def mock_code_file_manager(_mock_code_file_manager):
    yield _mock_code_file_manager


### Auto-used fixtures


def run_git_command(cwd, *args):
    """Helper function to run a git command."""
    subprocess.run(
        ["git"] + list(args),
        cwd=cwd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


@pytest.fixture(autouse=True)
def temp_testbed(monkeypatch):
    # create temporary copy of testbed, complete with git repo
    # realpath() resolves symlinks, required for paths to match on macOS
    temp_dir = os.path.realpath(tempfile.mkdtemp())
    temp_testbed = os.path.join(temp_dir, "testbed")
    shutil.copytree("testbed", temp_testbed)
    shutil.copy(".gitignore", temp_testbed)

    # Initialize git repo
    run_git_command(temp_testbed, "init")

    # Set local config for user.name and user.email. Set automatically on
    # MacOS, but not Windows/Ubuntu, which prevents commits from taking.
    run_git_command(temp_testbed, "config", "user.email", "test@example.com")
    run_git_command(temp_testbed, "config", "user.name", "Test User")

    # Add all files and commit
    run_git_command(temp_testbed, "add", ".")
    run_git_command(temp_testbed, "commit", "-m", "add testbed")

    # necessary to undo chdir before calling rmtree, or it fails on windows
    with monkeypatch.context() as m:
        m.chdir(temp_testbed)
        yield temp_testbed

    shutil.rmtree(temp_dir, onerror=add_permissions)


# Always set the user config to just be a config in the temp_testbed; that way,
# it will be unset unless a specific test wants to make a config in the testbed
@pytest.fixture(autouse=True)
def mock_user_config(mocker):
    from mentat import config_manager
    from mentat.config_manager import config_file_name

    config_manager.user_config_path = Path(config_file_name)


@pytest.fixture(autouse=True)
def mock_sleep_time(mocker):
    mocker.patch.object(StreamingPrinter, "sleep_time", new=lambda self: 0)



import fire
from operations import add_numbers, divide_numbers, multiply_numbers, subtract_numbers


def calculate(operation, num1, num2):
    result = None

    if operation == "add":
        result = add_numbers(num1, num2)
    elif operation == "subtract":
        result = subtract_numbers(num1, num2)
    elif operation == "multiply":
        result = multiply_numbers(num1, num2)
    elif operation == "divide":
        result = divide_numbers(num1, num2)
    else:
        print("Invalid operation")

    return result


if __name__ == "__main__":
    fire.Fire(calculate)



from __future__ import annotations

import glob
import logging
import os
from contextvars import ContextVar
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Optional

import attr

from .code_file import CodeFile
from .code_map import CodeMap
from .config_manager import CONFIG_MANAGER, ConfigManager
from .diff_context import DiffContext
from .errors import MentatError, UserError
from .git_handler import GIT_ROOT, get_non_gitignored_files, get_paths_with_git_diffs
from .llm_api import count_tokens, model_context_size
from .session_stream import SESSION_STREAM


def _is_file_text_encoded(file_path: Path):
    try:
        # The ultimate filetype test
        with open(file_path, "r") as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False


def _abs_files_from_list(paths: list[Path], check_for_text: bool = True):
    files_direct = set[CodeFile]()
    file_paths_from_dirs = set[Path]()
    for path in paths:
        file = CodeFile(os.path.realpath(path))
        path = Path(file.path)
        if path.is_file():
            if check_for_text and not _is_file_text_encoded(path):
                logging.info(f"File path {path} is not text encoded.")
                raise UserError(f"File path {path} is not text encoded.")
            files_direct.add(file)
        elif path.is_dir():
            nonignored_files = set(
                map(
                    lambda f: Path(os.path.realpath(path / f)),
                    get_non_gitignored_files(path),
                )
            )

            file_paths_from_dirs.update(
                filter(
                    lambda f: (not check_for_text) or _is_file_text_encoded(f),
                    nonignored_files,
                )
            )

    files_from_dirs = [
        CodeFile(os.path.realpath(path)) for path in file_paths_from_dirs
    ]
    return files_direct, files_from_dirs


def _get_files(
    config: ConfigManager, git_root: Path, paths: list[Path], exclude_paths: list[Path]
) -> Dict[Path, CodeFile]:
    excluded_files_direct, excluded_files_from_dirs = _abs_files_from_list(
        exclude_paths, check_for_text=False
    )
    excluded_files, excluded_files_from_dir = set(
        map(lambda f: f.path, excluded_files_direct)
    ), set(map(lambda f: f.path, excluded_files_from_dirs))

    glob_excluded_files = set(
        os.path.join(git_root, file)
        for glob_path in config.file_exclude_glob_list()
        # If the user puts a / at the beginning, it will try to look in root directory
        for file in glob.glob(
            pathname=glob_path,
            root_dir=git_root,
            recursive=True,
        )
    )
    files_direct, files_from_dirs = _abs_files_from_list(paths, check_for_text=True)

    # config glob excluded files only apply to files added from directories
    files_from_dirs = [
        file
        for file in files_from_dirs
        if str(file.path.resolve()) not in glob_excluded_files
    ]

    files_direct.update(files_from_dirs)

    files = dict[Path, CodeFile]()
    for file in files_direct:
        if file.path not in excluded_files | excluded_files_from_dir:
            files[Path(os.path.realpath(file.path))] = file

    return files


def _build_path_tree(files: list[CodeFile], git_root: Path):
    tree = dict[str, Any]()
    for file in files:
        path = os.path.relpath(file.path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


async def _print_path_tree(
    tree: dict[str, Any], changed_files: set[Path], cur_path: Path, prefix: str = ""
):
    stream = SESSION_STREAM.get()
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            await stream.send(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            await stream.send(f"{prefix}└── ", end="")

        cur = cur_path / key
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif star:
            color = "green"
        else:
            color = None
        await stream.send(f"{star}{key}", color=color)
        if tree[key]:
            await _print_path_tree(tree[key], changed_files, cur, new_prefix)


@attr.define
class CodeContextSettings:
    paths: list[Path] = []
    exclude_paths: list[Path] = []
    diff: Optional[str] = None
    pr_diff: Optional[str] = None
    no_code_map: bool = False

    def asdict(self):
        return attr.asdict(self)


CODE_CONTEXT: ContextVar[CodeContext] = ContextVar("mentat:code_context")


class CodeContext:
    settings: CodeContextSettings
    files: dict[Path, CodeFile]  # included path -> CodeFile
    file_lines: dict[Path, list[str]]  # included path -> cached lines

    def __init__(self, settings: CodeContextSettings):
        self.settings = settings

    @classmethod
    async def create(cls, settings: CodeContextSettings):
        self = cls(settings)
        await self.refresh(replace_paths=True)

        return self

    async def refresh(self, replace_paths: bool = False):
        stream = SESSION_STREAM.get()
        config = CONFIG_MANAGER.get()
        git_root = GIT_ROOT.get()

        # Diff context
        try:
            self.diff_context = DiffContext.create(
                self.settings.diff, self.settings.pr_diff
            )
            if replace_paths and not self.settings.paths:
                self.settings.paths = self.diff_context.files
        except UserError as e:
            await stream.send(str(e), color="light_yellow")
            raise MentatError("Failed to create diff context.")
        # User-specified Files
        self.files = _get_files(
            config, git_root, self.settings.paths, self.settings.exclude_paths
        )
        # Universal ctags
        self.code_map = (
            await CodeMap.create(token_limit=2048)
            if not self.settings.no_code_map
            else None
        )
        if self.code_map is not None and self.code_map.ctags_disabled:
            ctags_disabled_message = f"""
                There was an error with your universal ctags installation, disabling CodeMap.
                Reason: {self.code_map.ctags_disabled_reason}
            """
            ctags_disabled_message = dedent(ctags_disabled_message)
            await stream.send(ctags_disabled_message, color="yellow")

    async def display_context(self):
        stream = SESSION_STREAM.get()
        git_root = GIT_ROOT.get()

        if self.files:
            await stream.send("Files included in context:", color="green")
        else:
            await stream.send("No files included in context.\n", color="red")
            await stream.send("Git project: ", color="green", end="")
        await stream.send(git_root.name, color="blue")
        await _print_path_tree(
            _build_path_tree(list(self.files.values()), git_root),
            get_paths_with_git_diffs(),
            git_root,
        )
        await self.diff_context.display_context()

    async def get_code_message(
        self,
        all_file_lines: dict[Path, list[str]],
        model: str,
        provide_line_numbers: bool,
    ) -> str:
        stream = SESSION_STREAM.get()
        git_root = GIT_ROOT.get()

        code_message: list[str] = []
        if self.diff_context.files:
            code_message += [
                "Diff References:",
                f' "-" = {self.diff_context.name}',
                ' "+" = Active Changes',
                "",
            ]

        code_message += ["Code Files:\n"]
        for file in self.files.values():
            file_message: list[str] = []
            abs_path = file.path
            rel_path = Path(os.path.relpath(abs_path, git_root))

            # We always want to give GPT posix paths
            posix_rel_path = Path(rel_path).as_posix()
            file_message.append(posix_rel_path)

            file_lines = all_file_lines[rel_path]
            for i, line in enumerate(file_lines, start=1):
                if file.contains_line(i):
                    if provide_line_numbers:
                        file_message.append(f"{i}:{line}")
                    else:
                        file_message.append(f"{line}")
            file_message.append("")

            if rel_path in self.diff_context.files:
                file_message = self.diff_context.annotate_file_message(
                    rel_path, file_message
                )

            code_message += file_message

        if self.code_map is not None:
            code_message_tokens = count_tokens("\n".join(code_message), model)
            context_size = model_context_size(model)
            if context_size:
                max_tokens_for_code_map = context_size - code_message_tokens
                if self.code_map.token_limit:
                    code_map_message_token_limit = min(
                        self.code_map.token_limit, max_tokens_for_code_map
                    )
                else:
                    code_map_message_token_limit = max_tokens_for_code_map
            else:
                code_map_message_token_limit = self.code_map.token_limit

            code_map_message = await self.code_map.get_message(
                token_limit=code_map_message_token_limit
            )
            if code_map_message:
                match (code_map_message.level):
                    case "signatures":
                        level_message = "full syntax tree"
                    case "no_signatures":
                        level_message = "partial syntax tree"
                    case "filenames":
                        level_message = "filepaths only"

                message = f"\nIncluding CodeMap ({level_message})"
                await stream.send(message, color="green")
                code_message += [f"\n{code_map_message.content}"]
            else:
                message = [
                    "\nExcluding CodeMap from system message.",
                    "Reason: not enough tokens available in model context.",
                ]
                message = "\n".join(message)
                await stream.send(message, color="yellow")

        return "\n".join(code_message)

    async def add_file(self, code_file: CodeFile):
        stream = SESSION_STREAM.get()

        if not os.path.exists(code_file.path):
            await stream.send(f"File does not exist: {code_file.path}\n", color="red")
            return
        if code_file.path in self.files:
            await stream.send(
                f"File already in context: {code_file.path}\n", color="yellow"
            )
            return
        self.files[code_file.path] = code_file
        await stream.send(f"File added to context: {code_file.path}\n", color="green")

    async def remove_file(self, code_file: CodeFile):
        stream = SESSION_STREAM.get()

        if not os.path.exists(code_file.path):
            await stream.send(f"File does not exist: {code_file.path}\n", color="red")
            return
        if code_file.path not in self.files:
            await stream.send(
                f"File not in context: {code_file.path}\n", color="yellow"
            )
            return
        del self.files[code_file.path]
        await stream.send(
            f"File removed from context: {code_file.path}\n", color="green"
        )



import asyncio
import logging
import shlex
import traceback
from pathlib import Path
from typing import List, Optional, Union, cast
from uuid import uuid4

from mentat.logging_config import setup_logging

from .code_context import CODE_CONTEXT, CodeContext, CodeContextSettings
from .code_edit_feedback import get_user_feedback_on_edits
from .code_file_manager import CODE_FILE_MANAGER, CodeFileManager
from .commands import Command
from .config_manager import CONFIG_MANAGER, ConfigManager
from .conversation import CONVERSATION, Conversation
from .errors import MentatError
from .git_handler import GIT_ROOT, get_shared_git_root_for_paths
from .llm_api import COST_TRACKER, CostTracker, setup_api_key
from .parsers.block_parser import BlockParser
from .parsers.parser import PARSER, Parser
from .parsers.replacement_parser import ReplacementParser
from .parsers.split_diff_parser import SplitDiffParser
from .parsers.unified_diff_parser import UnifiedDiffParser
from .session_input import collect_user_input
from .session_stream import SESSION_STREAM, SessionStream

parser_map: dict[str, Parser] = {
    "block": BlockParser(),
    "replacement": ReplacementParser(),
    "split-diff": SplitDiffParser(),
    "unified-diff": UnifiedDiffParser(),
}


class Session:
    def __init__(
        self,
        stream: SessionStream,
    ):
        self.stream = stream

        self.id = uuid4()
        setup_api_key()

        self._main_task: asyncio.Task[None] | None = None
        self._stop_task: asyncio.Task[None] | None = None

    @classmethod
    async def create(
        cls,
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        no_code_map: bool = False,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
    ):
        # Set contextvars here
        stream = SessionStream()
        await stream.start()
        SESSION_STREAM.set(stream)

        cost_tracker = CostTracker()
        COST_TRACKER.set(cost_tracker)

        git_root = get_shared_git_root_for_paths([Path(path) for path in paths])
        GIT_ROOT.set(git_root)

        # TODO: Config should be created in the client (i.e., to get vscode settings) and passed to session
        config = await ConfigManager.create()
        CONFIG_MANAGER.set(config)

        parser = parser_map[config.parser()]
        PARSER.set(parser)

        code_context_settings = CodeContextSettings(
            paths, exclude_paths, diff, pr_diff, no_code_map
        )
        code_context = await CodeContext.create(code_context_settings)
        CODE_CONTEXT.set(code_context)

        # NOTE: Should codefilemanager, codecontext, and conversation be contextvars/singletons or regular instances?
        code_file_manager = CodeFileManager()
        CODE_FILE_MANAGER.set(code_file_manager)

        conversation = Conversation()
        CONVERSATION.set(conversation)

        return cls(stream)

    async def _main(self):
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()
        conversation = CONVERSATION.get()

        try:
            await code_context.display_context()
            await conversation.display_token_count()
        except MentatError as e:
            await stream.send(str(e), color="red")
            return

        await stream.send("Type 'q' or use Ctrl-C to quit at any time.", color="cyan")
        await stream.send("What can I do for you?", color="light_blue")
        need_user_request = True
        while True:
            if need_user_request:
                message = await collect_user_input()

                # Intercept and run command
                if isinstance(message.data, str) and message.data.startswith("/"):
                    arguments = shlex.split(message.data[1:])
                    command = Command.create_command(arguments[0])
                    await command.apply(*arguments[1:])
                    continue

                if message.data == "q":
                    break

                conversation.add_user_message(message.data)

            file_edits = await conversation.get_model_response()
            file_edits = [
                file_edit for file_edit in file_edits if await file_edit.is_valid()
            ]
            if file_edits:
                need_user_request = await get_user_feedback_on_edits(file_edits)
            else:
                need_user_request = True

    ### lifecycle

    @property
    def is_stopped(self):
        return self._main_task is None and self._stop_task is None

    def start(self) -> asyncio.Task[None]:
        """Asynchronously start the Session.

        A background asyncio. Task will be created to run the startup sequence and run
        the main loop which runs forever (until a client interrupts it).
        """

        if self._main_task:
            logging.warning("Job already started")
            return self._main_task

        setup_logging()

        async def run_main():
            try:
                await self._main()
                await cast(asyncio.Task[None], self.stop())
            except asyncio.CancelledError:
                pass

        def cleanup_main(task: asyncio.Task[None]):
            exception = task.exception()
            if exception is not None:
                logging.error(f"Main task for Session({self.id}) threw an exception")
                traceback.print_exception(
                    type(exception), exception, exception.__traceback__
                )

            self._main_task = None
            logging.debug("Main task stopped")

        self._main_task = asyncio.create_task(run_main())
        self._main_task.add_done_callback(cleanup_main)

        return self._main_task

    def stop(self) -> asyncio.Task[None] | None:
        """Asynchronously stop the Session.

        A background asyncio.Task will be created that handles the shutdown sequence
        of the Session. Clients should wait for `self.is_stopped` to return `True` in
        order to make sure the shutdown sequence has finished.
        """
        if self._stop_task is not None:
            logging.debug("Task is already stopping")
            return self._stop_task
        if self.is_stopped:
            logging.debug("Task is already stopped")
            return

        async def run_stop():
            cost_tracker = COST_TRACKER.get()

            if self._main_task is None:
                return
            try:
                await cost_tracker.display_total_cost()
                logging.shutdown()
                self._main_task.cancel()

                # Pyright can't see `self._main_task` being set to `None` in the task
                # callback handler, so we have to cast the type explicitly here
                self._main_task = cast(Union[asyncio.Task[None], None], self._main_task)

                while self._main_task is not None:
                    await asyncio.sleep(0.01)
                await self.stream.stop()
            except asyncio.CancelledError:
                pass

        def cleanup_stop(_: asyncio.Task[None]):
            self._stop_task = None
            logging.debug("Task has stopped")

        self._stop_task = asyncio.create_task(run_stop())
        self._stop_task.add_done_callback(cleanup_stop)

        return self._stop_task



import json
import subprocess

import fire

library_exceptions = [
    "mentat",
    # pip-licenses shows tiktoken's full license text, but it is MIT
    "tiktoken",
]
accepted_licenses = [
    "BSD License",
    "Apache Software License",
    "MIT License",
    "MIT",
    "Mozilla Public License 2.0 (MPL 2.0)",
    "Python Software Foundation License",
    "Apache 2.0",
    "BSD 3-Clause",
    "ISC License (ISCL)",
]


def main():
    licenses = json.loads(subprocess.check_output(["pip-licenses", "--format=json"]))
    for package in licenses:
        if "Name" not in package:
            raise ValueError(f"No name found for package {package}")
        elif "License" not in package:
            raise ValueError(f"License not found for package {package['Name']}")
        elif package["Name"] not in library_exceptions:
            package_licenses = package["License"].split(";")
            for package_license in package_licenses:
                if package_license.strip() not in accepted_licenses:
                    raise Exception(
                        f"Package {package['Name']} has license {package['License']};"
                        " if this license is valid, add it to the accepted licenses"
                        " list"
                    )
    print("It's an older license, sir, but it checks out")


if __name__ == "__main__":
    fire.Fire(main)



from __future__ import annotations

import json
import logging
from contextvars import ContextVar
from timeit import default_timer

from openai.error import InvalidRequestError, RateLimitError

from mentat.parsers.file_edit import FileEdit
from mentat.parsers.parser import PARSER, Parser
from tests.conftest import CODE_FILE_MANAGER, SessionStream

from .code_context import CODE_CONTEXT
from .config_manager import CONFIG_MANAGER, user_config_path
from .errors import MentatError, UserError
from .llm_api import (
    COST_TRACKER,
    call_llm_api,
    count_tokens,
    get_prompt_token_count,
    is_model_available,
    model_context_size,
)
from .session_stream import SESSION_STREAM

CONVERSATION: ContextVar[Conversation] = ContextVar("mentat:conversation")


class Conversation:
    def __init__(self):
        config = CONFIG_MANAGER.get()
        parser = PARSER.get()

        self.model = config.model()
        self.messages = list[dict[str, str]]()

        prompt = parser.get_system_prompt()
        self.add_system_message(prompt)

    async def display_token_count(self):
        stream = SESSION_STREAM.get()
        parser = PARSER.get()
        code_context = CODE_CONTEXT.get()
        config = CONFIG_MANAGER.get()
        code_file_manager = CODE_FILE_MANAGER.get()

        if not is_model_available(self.model):
            raise MentatError(
                f"Model {self.model} is not available. Please try again with a"
                " different model."
            )
        if "gpt-4" not in self.model:
            await stream.send(
                "Warning: Mentat has only been tested on GPT-4. You may experience"
                " issues with quality. This model may not be able to respond in"
                " mentat's edit format.",
                color="yellow",
            )
            if "gpt-3.5" not in self.model:
                await stream.send(
                    "Warning: Mentat does not know how to calculate costs or context"
                    " size for this model.",
                    color="yellow",
                )

        prompt = parser.get_system_prompt()
        code_file_manager.read_all_file_lines()
        tokens = count_tokens(
            await code_context.get_code_message(
                code_file_manager.file_lines, self.model, parser.provide_line_numbers()
            ),
            self.model,
        ) + count_tokens(prompt, self.model)
        context_size = model_context_size(self.model)
        maximum_context = config.maximum_context()
        if maximum_context:
            if context_size:
                context_size = min(context_size, maximum_context)
            else:
                context_size = maximum_context

        if not context_size:
            raise MentatError(
                f"Context size for {self.model} is not known. Please set"
                f" maximum-context in {user_config_path}."
            )
        if context_size and tokens > context_size:
            raise MentatError(
                f"Included files already exceed token limit ({tokens} /"
                f" {context_size}). Please try running again with a reduced"
                " number of files."
            )
        elif tokens + 1000 > context_size:
            await stream.send(
                f"Warning: Included files are close to token limit ({tokens} /"
                f" {context_size}), you may not be able to have a long"
                " conversation.",
                color="red",
            )
        else:
            await stream.send(
                f"File and prompt token count: {tokens} / {context_size}",
                color="cyan",
            )

    def add_system_message(self, message: str):
        self.messages.append({"role": "system", "content": message})

    def add_user_message(self, message: str):
        self.messages.append({"role": "user", "content": message})

    def add_assistant_message(self, message: str):
        self.messages.append({"role": "assistant", "content": message})

    async def _stream_model_response(
        self,
        stream: SessionStream,
        parser: Parser,
        messages: list[dict[str, str]],
    ):
        start_time = default_timer()
        try:
            response = await call_llm_api(messages, self.model)
            await stream.send(
                "Streaming... use control-c to interrupt the model at any point\n"
            )
            async with parser.interrupt_catcher():
                message, file_edits = await parser.stream_and_parse_llm_response(
                    response
                )
        except InvalidRequestError as e:
            raise MentatError(
                "Something went wrong - invalid request to OpenAI API. OpenAI"
                " returned:\n"
                + str(e)
            )
        except RateLimitError as e:
            raise UserError("OpenAI gave a rate limit error:\n" + str(e))

        time_elapsed = default_timer() - start_time
        return (message, file_edits, time_elapsed)

    async def get_model_response(self) -> list[FileEdit]:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()
        cost_tracker = COST_TRACKER.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        parser = PARSER.get()

        messages = self.messages.copy()

        code_file_manager.read_all_file_lines()
        code_message = await code_context.get_code_message(
            code_file_manager.file_lines, self.model, parser.provide_line_numbers()
        )
        messages.append({"role": "system", "content": code_message})

        num_prompt_tokens = await get_prompt_token_count(messages, self.model)
        message, file_edits, time_elapsed = await self._stream_model_response(
            stream, parser, messages
        )
        await cost_tracker.display_api_call_stats(
            num_prompt_tokens,
            count_tokens(message, self.model),
            self.model,
            time_elapsed,
        )

        transcript_logger = logging.getLogger("transcript")
        messages.append({"role": "assistant", "content": message})
        transcript_logger.info(json.dumps({"messages": messages}))

        self.add_assistant_message(message)
        return file_edits



# Raised when model output doesn't adhere to the specified format for changes
class ModelError(Exception):
    pass


# Used to indicate an issue with Mentat's code
class MentatError(Exception):
    pass


# Used to indicate an issue with the user's usage of Mentat
class UserError(Exception):
    pass


# Used to indicate a KeyboardInterupt thrown by a remote client
class RemoteKeyboardInterrupt(Exception):
    pass



import os
import shutil
import subprocess
from textwrap import dedent

import pytest

from mentat.session import Session

# These benchmarks use GPT and won't run automatically.
# Run them with python tests/record_benchmark.py True
# to record the results or use pytest --benchmark
# to not save the results.
pytestmark = pytest.mark.benchmark


async def edit_file_and_run(
    mock_collect_user_input, prompts, context_file_paths, main_file_path, argument_lists
):
    mock_collect_user_input.set_stream_messages(
        [prompt for pair in zip(prompts, ["y"] * len(prompts)) for prompt in pair]
        + ["q"]
    )

    session = await Session.create(context_file_paths)
    await session.start()
    await session.stream.stop()

    assert os.path.exists(main_file_path)

    results = []
    for arguments in argument_lists:
        result = subprocess.check_output(
            ["python", main_file_path, *arguments], text=True
        )
        results.append(result.strip())
    return results


@pytest.mark.asyncio
async def test_calculator_add_power(mock_collect_user_input):
    calculator_path = "scripts/calculator.py"
    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=[
            "Add power as a possible operation, raising the first arg to the power of"
            " the second"
        ],
        context_file_paths=[calculator_path],
        main_file_path=calculator_path,
        argument_lists=[["power", "15", "3"]],
    )
    assert float(results[0]) == 3375.0


@pytest.mark.asyncio
async def test_calculator_add_exp_then_log(mock_collect_user_input):
    calculator_path = "scripts/calculator.py"
    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=[
            "Add exponentation operation, called with 'exp'",
            "Add logarithm operation, called with 'log'",
        ],
        context_file_paths=[calculator_path],
        main_file_path=calculator_path,
        argument_lists=[["exp", "15", "3"], ["log", "10", "2"]],
    )
    assert float(results[0]) == 3375.0
    assert float(results[1]) == 3.3219280948873626


@pytest.mark.asyncio
async def test_calculator_continue_change(mock_collect_user_input):
    calculator_path = "scripts/calculator.py"

    with open(calculator_path, "r") as f:
        calculator_lines = f.readlines()
    index = calculator_lines.index("    return a / b\n")
    new_lines = [
        "\n",
        "\n",
        "def exp_numbers(a, b):\n",
        "    return a**b\n",
    ]
    calculator_lines = (
        calculator_lines[: index + 1] + new_lines + calculator_lines[index + 1 :]
    )
    with open(calculator_path, "w") as f:
        f.writelines(calculator_lines)

    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=["complete the change I started"],
        context_file_paths=[calculator_path],
        main_file_path=calculator_path,
        argument_lists=[["exp", "15", "3"]],
    )
    assert float(results[0]) == 3375.0


@pytest.mark.asyncio
async def test_multifile_calculator(mock_collect_user_input):
    multifile_calculator_path = "multifile_calculator"
    calculator_path = os.path.join(multifile_calculator_path, "calculator.py")
    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=["add exp and log functions to take a^b and log a base b"],
        context_file_paths=[multifile_calculator_path],
        main_file_path=calculator_path,
        argument_lists=[["exp", "15", "3"], ["log", "10", "2"]],
    )

    assert float(results[0]) == 3375.0
    assert float(results[1]) == 3.3219280948873626


@pytest.mark.asyncio
async def test_start_project_from_scratch(mock_collect_user_input):
    # Clear the testbed so we can test that it works with empty directories
    for item in os.listdir("."):
        if os.path.isfile(item):
            os.remove(item)
        elif os.path.isdir(item):
            if item != ".git":
                shutil.rmtree(item)
    subprocess.run(["git", "rm", "-r", "--cached", "."])

    fizzbuzz_path = "fizzbuzz.py"
    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=["make a file that does fizzbuzz, named fizzbuzz.py, going up to 10"],
        context_file_paths=["."],
        main_file_path=fizzbuzz_path,
        argument_lists=[[]],
    )

    expected_output = "1\n2\nFizz\n4\nBuzz\nFizz\n7\n8\nFizz\nBuzz"
    assert results[0] == expected_output


@pytest.mark.asyncio
async def test_import_from_code_map(mock_collect_user_input):
    prompt = """
        make a file called encoded_echo.py that gets the base64 encoded value of
        `echo_hardcoded` and print the encoded value as a string
    """
    prompt = dedent(prompt).strip()
    encoded_echo_path = "encoded_echo.py"
    results = await edit_file_and_run(
        mock_collect_user_input,
        prompts=[prompt],
        context_file_paths=[],
        main_file_path=encoded_echo_path,
        argument_lists=[[]],
    )

    expected_output = "MDBtcDk0ITJQZ0slaTJoQG0mNVh6azRUJlZxQypOV1FebXA5NCEyUGdLJWkyaEBtJjVYems0VCZWcUMqTldRXg=="
    assert results[0] == expected_output



import asyncio


async def run_subprocess_async(*args: str) -> str:
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        start_new_session=True,
    )
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        error_output = stderr.decode("utf-8").strip() if stderr else ""
        raise Exception(f"Subprocess failed with error: {error_output}")

    output = stdout.decode("utf-8").strip() if stdout else ""

    return output



from typing import Any

from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import FormattedText

from mentat.session_stream import StreamMessage


def _print_stream_message_string(
    content: Any,
    end: str = "\n",
    color: str | None = None,
    flush: bool = False,
):
    """
    Do NOT mix termcolor colored with prompt_toolkit FormattedText! If colored text gets sent here and
    color is not None, the FormattedText will undo the colored text and display all of the ANSI codes.
    """
    if color is not None:
        f_color = color.replace("_", "").replace("light", "bright")
        if f_color != "" and not f_color.startswith("ansi"):
            f_color = "ansi" + f_color
        print_formatted_text(FormattedText([(f_color, content)]), end=end, flush=flush)
    else:
        print(content, end=end, flush=True)


def print_stream_message(message: StreamMessage):
    end = "\n"
    color = None
    flush = False
    if message.extra:
        if isinstance(message.extra.get("end"), str):
            end = message.extra["end"]
        if isinstance(message.extra.get("color"), str):
            color = message.extra["color"]
        if isinstance(message.extra.get("flush"), bool):
            flush = message.extra["flush"]

    _print_stream_message_string(
        content=message.data,
        end=end,
        color=color,
        flush=flush,
    )



from __future__ import annotations

import asyncio
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, AsyncGenerator, Dict, List, cast
from uuid import UUID, uuid4

from .broadcast import Broadcast

SESSION_STREAM: ContextVar[SessionStream] = ContextVar("mentat:session_stream")


class StreamMessageSource(Enum):
    SERVER = "server"
    CLIENT = "client"


@dataclass(slots=True)
class StreamMessage:
    id: UUID
    channel: str
    source: StreamMessageSource
    data: Any
    extra: Dict[str, Any] | None
    created_at: datetime


class SessionStream:
    """Replaces `cprint` and `print`

    Stores message history for a Session and holds an in-memory message bus.

    Terminal and extension clients can read these messages and render them accordingly.
    For the terminal, they would be rendered with `cprint`.
    """

    def __init__(self):
        self.messages: List[StreamMessage] = []
        self.interrupt_lock = asyncio.Lock()
        self._broadcast = Broadcast()

    async def start(self):
        await self._broadcast.connect()

    async def stop(self):
        await self._broadcast.disconnect()

    async def send(
        self,
        data: Any,
        source: StreamMessageSource = StreamMessageSource.SERVER,
        channel: str = "default",
        **kwargs: Any,
    ):
        message = StreamMessage(
            id=uuid4(),
            source=source,
            channel=channel,
            data=data,
            created_at=datetime.utcnow(),
            extra=kwargs,
        )

        self.messages.append(message)
        await self._broadcast.publish(channel=channel, message=message)

        return message

    async def recv(self, channel: str = "default") -> StreamMessage:
        """Listen for a single event on a channel"""
        async with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                stream_message = cast(StreamMessage, event.message)
                return stream_message
            raise Exception("recv should not complete without receiving an Event")

    async def listen(
        self, channel: str = "default"
    ) -> AsyncGenerator[StreamMessage, None]:
        """Listen to all messages on a channel indefinitely"""
        async with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                yield event.message



import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

from .errors import UserError
from .git_handler import (
    GIT_ROOT,
    check_head_exists,
    get_diff_for_file,
    get_files_in_diff,
    get_treeish_metadata,
)
from .session_stream import SESSION_STREAM


@dataclass
class DiffAnnotation:
    start: int
    message: list[str]

    @property
    def length(self):
        return sum(bool(line.startswith("+")) for line in self.message)


def _parse_diff(diff: str) -> list[DiffAnnotation]:
    """Parse diff into a list of annotations."""
    annotations: list[DiffAnnotation] = []
    active_annotation: Optional[DiffAnnotation] = None
    lines = diff.splitlines()
    for line in lines[4:]:  # Ignore header
        if line.startswith(("---", "+++", "//")):
            continue
        elif line.startswith("@@"):
            if active_annotation:
                annotations.append(active_annotation)
            _new_index = line.split(" ")[2]
            if "," in _new_index:
                new_start = _new_index[1:].split(",")[0]
            else:
                new_start = _new_index[1:]
            active_annotation = DiffAnnotation(int(new_start), [])
        elif line.startswith(("+", "-")):
            if not active_annotation:
                raise UserError("Invalid diff")
            active_annotation.message.append(line)
    if active_annotation:
        annotations.append(active_annotation)
    annotations.sort(key=lambda a: a.start)
    return annotations


def _annotate_file_message(
    code_message: list[str], annotations: list[DiffAnnotation]
) -> list[str]:
    """Return the code_message with annotations inserted."""
    active_index = 0
    annotated_message: list[str] = []
    for annotation in annotations:
        # Fill-in lines between annotations
        if active_index < annotation.start:
            unaffected_lines = code_message[active_index : annotation.start]
            annotated_message += unaffected_lines
        active_index = annotation.start
        if annotation.start == 0:
            # Make sure the PATH stays on line 1
            annotated_message.append(code_message[0])
            active_index += 1
        i_minus = None
        for line in annotation.message:
            sign = line[0]
            if sign == "+":
                # Add '+' lines in place of code_message lines
                annotated_message.append(f"{active_index}:{line}")
                active_index += 1
                i_minus = None
            elif sign == "-":
                # Insert '-' lines at the point they were removed
                i_minus = 0 if i_minus is None else i_minus
                annotated_message.append(f"{annotation.start + i_minus}:{line}")
                i_minus += 1
    if active_index < len(code_message):
        annotated_message += code_message[active_index:]

    return annotated_message


class DiffContext:
    def __init__(
        self,
        target: Optional[str] = None,
        name: Optional[str] = None,
    ):
        if target is None:
            self.target = "HEAD"
            self.name = "HEAD (last commit)"
        else:
            self.target = target
            self.name = name

    @classmethod
    def create(
        cls,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
    ):
        git_root = GIT_ROOT.get()

        if diff and pr_diff:
            raise UserError("Cannot specify more than one type of diff.")

        target = diff or pr_diff
        if not target:
            return cls()

        name = ""
        treeish_type = _get_treeish_type(git_root, target)
        if treeish_type == "branch":
            name += f"Branch {target}: "
        elif treeish_type == "relative":
            name += f"{target}: "

        if pr_diff:
            name = f"Merge-base {name}"
            target = _git_command(git_root, "merge-base", "HEAD", pr_diff)
            if not target:
                raise UserError(
                    f"Cannot identify merge base between HEAD and {pr_diff}"
                )

        meta = get_treeish_metadata(target)
        name += f'{meta["hexsha"][:8]}: {meta["summary"]}'
        return cls(target, name)

    @property
    def files(self) -> list[Path]:
        if self.target == "HEAD" and not check_head_exists():
            return []  # A new repo without any commits
        return get_files_in_diff(self.target)

    async def display_context(self) -> None:
        stream = SESSION_STREAM.get()

        if not self.files:
            return
        await stream.send("Diff annotations:", color="green")
        num_files = len(self.files)
        num_lines = 0
        # TODO: Only include paths in context
        for file in self.files:
            diff = get_diff_for_file(self.target, file)
            diff_lines = diff.splitlines()
            num_lines += len(
                [line for line in diff_lines if line.startswith(("+ ", "- "))]
            )
        await stream.send(f" ─•─ {self.name} | {num_files} files | {num_lines} lines\n")

    def annotate_file_message(
        self, rel_path: Path, file_message: list[str]
    ) -> list[str]:
        """Return file_message annotated with active diff."""

        if not self.files:
            return file_message
        diff = get_diff_for_file(self.target, rel_path)
        annotations = _parse_diff(diff)
        return _annotate_file_message(file_message, annotations)


TreeishType = Literal["commit", "branch", "relative"]


def _git_command(git_root: Path, *args: str) -> str | None:
    try:
        return subprocess.check_output(
            ["git"] + list(args), cwd=git_root, stderr=subprocess.PIPE, text=True
        ).strip()
    except subprocess.CalledProcessError:
        return None


def _get_treeish_type(git_root: Path, treeish: str) -> TreeishType:
    object_type = _git_command(git_root, "cat-file", "-t", treeish)

    if not object_type:
        raise UserError(f"Invalid treeish: {treeish}")

    if object_type == "commit":
        if "~" in treeish or "^" in treeish:
            return "relative"

        if _git_command(git_root, "show-ref", "--heads", treeish):
            return "branch"
        else:
            return "commit"

    raise UserError(f"Unsupported treeish type: {object_type}")



import os
from pathlib import Path

from setuptools import find_packages, setup


def read_requirements(file):
    with open(file, "r") as f:
        return f.read().splitlines()


readme_path = os.path.join(Path(__file__).parent, "README.md")
with open(readme_path, "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="mentat",
    version="0.1.15",
    python_requires=">=3.10",
    packages=find_packages(),
    install_requires=read_requirements("requirements.txt"),
    package_data={
        "mentat": ["default_config.json", "prompts/*.txt"],
    },
    entry_points={
        "console_scripts": [
            "mentat=mentat.terminal.client:run_cli",
        ],
    },
    description="AI coding assistant on your command line",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
)



import os
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.session import Session


# Make sure we always give posix paths to GPT
@pytest.mark.asyncio
async def test_posix_paths(
    mock_stream, mock_config, mock_code_file_manager, mock_code_context
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = Path(os.path.join(dir_name, file_name))
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")

    mock_code_context.settings.paths = [file_path]
    await mock_code_context.refresh(True)

    mock_code_file_manager.read_all_file_lines()
    code_message = await mock_code_context.get_code_message(
        mock_code_file_manager.file_lines, mock_config.model(), True
    )
    assert dir_name + "/" + file_name in code_message.split("\n")


@pytest.mark.asyncio
async def test_partial_files(
    mock_stream, mock_config, mock_code_file_manager, mock_code_context
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = os.path.join(dir_name, file_name)
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write(dedent("""\
             I am a file
             with 5 lines
             third
             fourth
             fifth"""))

    file_path_partial = file_path + ":1,3-5"
    mock_code_context.settings.paths = [Path(file_path_partial)]
    await mock_code_context.refresh(True)
    mock_code_context.code_map = None

    mock_code_file_manager.read_all_file_lines()
    code_message = await mock_code_context.get_code_message(
        mock_code_file_manager.file_lines, mock_config.model(), True
    )
    assert code_message == dedent("""\
            Code Files:

            dir/file.txt
            1:I am a file
            3:third
            4:fourth
            5:fifth
              """)


@pytest.mark.asyncio
async def test_run_from_subdirectory(
    mock_collect_user_input, mock_call_llm_api, mock_setup_api_key
):
    """Run mentat from a subdirectory of the git root"""
    # Change to the subdirectory
    os.chdir("multifile_calculator")
    mock_collect_user_input.set_stream_messages(
        [
            (
                "Insert the comment # Hello on the first line of"
                " multifile_calculator/calculator.py and scripts/echo.py"
            ),
            "y",
            "q",
        ]
    )
    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment in both files.

        @@start
        {
            "file": "multifile_calculator/calculator.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end
        @@start
        {
            "file": "scripts/echo.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end""")])

    session = await Session.create([Path("calculator.py"), Path("../scripts")])
    await session.start()
    await session.stream.stop()

    # Check that it works
    with open("calculator.py") as f:
        calculator_output = f.readlines()
    with open("../scripts/echo.py") as f:
        echo_output = f.readlines()
    assert calculator_output[0].strip() == "# Hello"
    assert echo_output[0].strip() == "# Hello"



import json
import tempfile
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Any, Literal

from .config_manager import CONFIG_MANAGER, ConfigManager
from .git_handler import GIT_ROOT, get_non_gitignored_files
from .llm_api import count_tokens
from .session_stream import SESSION_STREAM
from .utils import run_subprocess_async


async def _get_code_map(
    root: Path, file_path: Path, exclude_signatures: bool = False
) -> str:
    # Create ctags from executable in a subprocess
    ctags_cmd_args = [
        "--extras=-F",
        "--input-encoding=utf-8",
        "--output-format=json",
        "--output-encoding=utf-8",
    ]
    if exclude_signatures:
        ctags_cmd_args.append("--fields=-s")
    else:
        ctags_cmd_args.append("--fields=+S")
    ctags_cmd = ["ctags", *ctags_cmd_args, str(Path(root).joinpath(file_path))]

    output = await run_subprocess_async(*ctags_cmd)
    output_lines = output.splitlines()

    # Extract subprocess stdout into python objects
    ctags = set[tuple[Path, ...]]()
    for output_line in output_lines:
        try:
            tag = json.loads(output_line)
        except json.decoder.JSONDecodeError as err:
            await SESSION_STREAM.get().send(
                f"Error parsing ctags output: {err}\n{repr(output_line)}",
                color="yellow",
            )
            continue

        scope = tag.get("scope")
        kind = tag.get("kind")
        name = tag.get("name")
        signature = tag.get("signature")

        last = name
        if signature:
            last += " " + signature

        res = [file_path]
        if scope:
            res.append(scope)
        res += [kind, last]

        ctags.add(tuple(res))

    if len(ctags) == 0:
        return f"{file_path}\n"

    # Build LLM-readable string representation of ctag objects
    sorted_tags = sorted(ctags)
    output = ""
    last = [None] * len(sorted_tags[0])
    tab = "\t"
    for tag in sorted_tags:
        tag = list(tag)

        num_common = 0
        for i in range(len(last) + 1):
            if i == len(last):
                break
            if last[i] != tag[i]:
                num_common = i
                break

        indent = tab * num_common
        rest = tag[num_common:]

        for item in rest:
            output += indent + str(item) + "\n"
            indent += tab
        last = tag

    return output


def _get_file_map(file_paths: set[Path]) -> str:
    tree = dict[str, Any]()
    for file_path in file_paths:
        parts = file_path.parts
        node = tree
        for part in parts:
            if part not in node:
                node[part] = {}
            node = node[part]

    def tree_to_string(tree: dict[str, Any], indent: int = 0) -> str:
        s = ""
        sorted_keys = sorted(tree.keys())
        for key in sorted_keys:
            s += "\t" * indent + key + "\n"
            s += tree_to_string(tree[key], indent + 1)
        return s

    file_map_tree = tree_to_string(tree)
    return file_map_tree


@dataclass
class CodeMapMessage:
    level: Literal["signatures", "no_signatures", "filenames"]
    content: str


class CodeMap:
    def __init__(self, token_limit: int | None = None):
        self.token_limit = token_limit
        self.ctags_disabled = True
        self.ctags_disabled_reason = ""

    @classmethod
    async def create(cls, token_limit: int | None = None):
        stream = SESSION_STREAM.get()

        self = cls(token_limit)
        await self._check_ctags_executable()
        if self.ctags_disabled:
            ctags_disabled_message = f"""
                There was an error with your universal ctags installation, disabling CodeMap.
                Reason: {self.ctags_disabled_reason}
            """
            ctags_disabled_message = dedent(ctags_disabled_message)
            await stream.send(ctags_disabled_message, color="yellow")

        return self

    async def _check_ctags_executable(self):
        try:
            cmd = ["ctags", "--version"]
            output = await run_subprocess_async(*cmd)
            output = output.lower()

            cmd = " ".join(cmd)
            if "universal ctags" not in output:
                self.ctags_disabled_reason = (
                    f"{cmd} does not claim to be universal ctags"
                )
                return
            if "+json" not in output:
                self.ctags_disabled_reason = f"{cmd} does not list +json support"
                return

            with tempfile.TemporaryDirectory() as tempdir:
                hello_py = Path(tempdir) / "hello.py"
                with open(hello_py, "w", encoding="utf-8") as f:
                    f.write("def hello():\n    print('Hello, world!')\n")
                await _get_code_map(Path(tempdir), hello_py)
        except FileNotFoundError:
            self.ctags_disabled_reason = "ctags executable not found"
            return
        except Exception as e:
            self.ctags_disabled_reason = f"error running universal-ctags: {e}"
            return

        self.ctags_disabled = False

    async def _get_message_from_ctags(
        self,
        config: ConfigManager,
        root: Path,
        file_paths: set[Path],
        exclude_signatures: bool = False,
        token_limit: int | None = None,
    ) -> CodeMapMessage | None:
        token_limit = token_limit or self.token_limit
        code_maps = list[str]()
        code_maps_token_count = 0
        for file_path in file_paths:
            code_map = await _get_code_map(
                root, file_path, exclude_signatures=exclude_signatures
            )
            code_map_token_count = count_tokens(code_map, config.model())

            if token_limit is not None and code_maps_token_count > token_limit:
                if exclude_signatures is True:
                    return
                return await self._get_message_from_ctags(
                    config,
                    root,
                    file_paths,
                    exclude_signatures=True,
                    token_limit=token_limit,
                )

            code_maps.append(code_map)
            code_maps_token_count += code_map_token_count

        message = "Code Map:" + "\n\n" + "\n".join(code_maps)

        message_token_count = count_tokens(message, config.model())
        if token_limit is not None and message_token_count > token_limit:
            if exclude_signatures is True:
                return
            return await self._get_message_from_ctags(
                config, root, file_paths, exclude_signatures=True
            )

        code_map_message = CodeMapMessage(
            level="signatures" if not exclude_signatures else "no_signatures",
            content=message,
        )

        return code_map_message

    def _get_message_from_file_map(
        self,
        config: ConfigManager,
        file_paths: set[Path],
        token_limit: int | None = None,
    ) -> CodeMapMessage | None:
        file_map_tree = _get_file_map(file_paths)
        message = "Code Map:" + "\n\n" + file_map_tree

        message_token_count = count_tokens(message, config.model())
        token_limit = token_limit or self.token_limit
        if token_limit is not None and message_token_count > token_limit:
            return

        code_map_message = CodeMapMessage(level="filenames", content=message)

        return code_map_message

    async def get_message(
        self, token_limit: int | None = None
    ) -> CodeMapMessage | None:
        config = CONFIG_MANAGER.get()
        git_root = GIT_ROOT.get()

        git_file_paths = get_non_gitignored_files(git_root)

        if not self.ctags_disabled:
            code_map_message = await self._get_message_from_ctags(
                config, git_root, git_file_paths, token_limit=token_limit
            )
            if code_map_message is not None:
                return code_map_message

        file_map_message = self._get_message_from_file_map(
            config, git_file_paths, token_limit=token_limit
        )
        if file_map_message is not None:
            return file_map_message



import argparse
import asyncio
import glob
import logging
import signal
from pathlib import Path
from typing import Any, Coroutine, List, Set

from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.styles import Style
from termcolor import cprint

from mentat.code_file import parse_intervals
from mentat.config_manager import CONFIG_MANAGER
from mentat.session import Session
from mentat.session_stream import StreamMessageSource
from mentat.terminal.output import print_stream_message
from mentat.terminal.prompt_completer import MentatCompleter
from mentat.terminal.prompt_session import MentatPromptSession


class TerminalClient:
    def __init__(
        self,
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        no_code_map: bool = False,
        diff: str | None = None,
        pr_diff: str | None = None,
    ):
        self.paths = paths
        self.exclude_paths = exclude_paths
        self.no_code_map = no_code_map
        self.diff = diff
        self.pr_diff = pr_diff

        self.session: Session | None = None

        self._tasks: Set[asyncio.Task[None]] = set()
        self._should_exit = False
        self._force_exit = False

    def _create_task(self, coro: Coroutine[None, None, Any]):
        """Utility method for running a Task in the background"""

        def task_cleanup(task: asyncio.Task[None]):
            self._tasks.remove(task)

        task = asyncio.create_task(coro)
        task.add_done_callback(task_cleanup)
        self._tasks.add(task)

        return task

    async def _cprint_session_stream(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"
        async for message in self.session.stream.listen():
            print_stream_message(message)

    async def _handle_input_requests(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"
        while True:
            input_request_message = await self.session.stream.recv("input_request")
            # TODO: Make extra kwargs like plain constants
            if (
                input_request_message.extra is not None
                and input_request_message.extra.get("plain")
            ):
                prompt_session = self._plain_session
            else:
                prompt_session = self._prompt_session
            user_input = await prompt_session.prompt_async(handle_sigint=False)
            assert isinstance(user_input, str)
            if user_input == "q":
                self._should_exit = True
                return

            await self.session.stream.send(
                user_input,
                source=StreamMessageSource.CLIENT,
                channel=f"input_request:{input_request_message.id}",
            )

    async def _send_session_stream_interrupt(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"
        await self.session.stream.send(
            "", source=StreamMessageSource.CLIENT, channel="interrupt"
        )

    def _handle_exit(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"
        if (
            self.session.is_stopped
            or self.session.stream.interrupt_lock.locked() is False
        ):
            if self._should_exit:
                logging.debug("Force exiting client...")
                self._force_exit = True
            else:
                logging.debug("Should exit client...")
                self._should_exit = True

        else:
            logging.debug("Sending interrupt to session stream")
            self._create_task(self._send_session_stream_interrupt())

    def _init_signal_handlers(self):
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, self._handle_exit)

    async def _startup(self):
        assert self.session is None, "TerminalClient already running"

        self.session = await Session.create(
            self.paths, self.exclude_paths, self.no_code_map, self.diff, self.pr_diff
        )
        self.session.start()
        # Logging is setup in session.start()
        logging.debug("Running startup")

        mentat_completer = MentatCompleter()
        self._prompt_session = MentatPromptSession(completer=mentat_completer)

        plain_bindings = KeyBindings()

        @plain_bindings.add("c-c")
        @plain_bindings.add("c-d")
        def _(event: KeyPressEvent):
            if event.current_buffer.text != "":
                event.current_buffer.reset()
            else:
                event.app.exit(result="q")

        self._plain_session = PromptSession[str](
            message=[("class:prompt", ">>> ")],
            style=Style(CONFIG_MANAGER.get().input_style()),
            completer=None,
            key_bindings=plain_bindings,
        )

        self._create_task(mentat_completer.refresh_completions())
        self._create_task(self._cprint_session_stream())
        self._create_task(self._handle_input_requests())

        logging.debug("Completed startup")

    async def _shutdown(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"

        logging.debug("Running shutdown")

        # Stop session
        self.session.stop()
        while not self._force_exit and not self.session.is_stopped:
            await asyncio.sleep(0.01)
        self.session = None
        # logging is shutdown by session stop

        # Stop all background tasks
        for task in self._tasks:
            task.cancel()
        while not self._force_exit:
            if all([task.cancelled() for task in self._tasks]):
                break
            await asyncio.sleep(0.01)

    async def _main(self):
        assert isinstance(self.session, Session), "TerminalClient is not running"
        logging.debug("Running main loop")

        while not self._should_exit and not self.session.is_stopped:
            await asyncio.sleep(0.01)

    async def _run(self):
        self._init_signal_handlers()
        await self._startup()
        await self._main()
        await self._shutdown()

    def run(self):
        asyncio.run(self._run())


def expand_paths(paths: list[str]) -> list[Path]:
    globbed_paths = set[str]()
    invalid_paths = list[str]()
    for path in paths:
        new_paths = glob.glob(pathname=path, recursive=True)
        if new_paths:
            globbed_paths.update(new_paths)
        else:
            split = path.rsplit(":", 1)
            p = split[0]
            if len(split) > 1:
                # Parse additional syntax, e.g. "path/to/file.py:1-5,7,12-40"
                intervals = parse_intervals(split[1])
            else:
                intervals = None
            if Path(p).exists() and intervals:
                globbed_paths.add(path)
            else:
                invalid_paths.append(path)
    if invalid_paths:
        cprint("The following paths do not exist:", "light_yellow")
        print("\n".join(invalid_paths))
        exit()
    return [Path(path) for path in globbed_paths]


def run_cli():
    parser = argparse.ArgumentParser(
        description="Run conversation with command line args"
    )
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns",
    )
    parser.add_argument(
        "--exclude",
        "-e",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns to exclude",
    )
    parser.add_argument(
        "--no-code-map",
        action="store_true",
        help="Exclude the file structure/syntax map from the system prompt",
    )
    parser.add_argument(
        "--diff",
        "-d",
        type=str,
        default=None,
        help="A git tree-ish (e.g. commit, branch, tag) to diff against",
    )
    parser.add_argument(
        "--pr-diff",
        "-p",
        type=str,
        default=None,
        help="A git tree-ish to diff against the latest common ancestor of",
    )
    args = parser.parse_args()
    paths = args.paths
    exclude_paths = args.exclude
    no_code_map = args.no_code_map
    diff = args.diff
    pr_diff = args.pr_diff

    # Expanding paths as soon as possible because some shells such as zsh automatically
    # expand globs and we want to avoid differences in functionality between shells
    terminal_client = TerminalClient(
        expand_paths(paths), expand_paths(exclude_paths), no_code_map, diff, pr_diff
    )
    terminal_client.run()



import logging
import os
import subprocess
from contextvars import ContextVar
from pathlib import Path

from mentat.errors import UserError

GIT_ROOT: ContextVar[Path] = ContextVar("mentat:git_root")


def get_git_diff_for_path(path: Path) -> str:
    git_root = GIT_ROOT.get()
    return subprocess.check_output(["git", "diff", path], cwd=git_root).decode("utf-8")


def get_non_gitignored_files(path: Path) -> set[Path]:
    return set(
        # git returns / separated paths even on windows, convert so we can remove
        # glob_excluded_files, which have windows paths on windows
        Path(os.path.normpath(p))
        for p in filter(
            lambda p: p != "",
            subprocess.check_output(
                # -c shows cached (regular) files, -o shows other (untracked/new) files
                ["git", "ls-files", "-c", "-o", "--exclude-standard"],
                cwd=path,
                text=True,
            ).split("\n"),
        )
    )


def get_paths_with_git_diffs() -> set[Path]:
    git_root = GIT_ROOT.get()

    changed = subprocess.check_output(
        ["git", "diff", "--name-only"], cwd=git_root, text=True
    ).split("\n")
    new = subprocess.check_output(
        ["git", "ls-files", "-o", "--exclude-standard"], cwd=git_root, text=True
    ).split("\n")
    return set(
        map(
            lambda path: Path(os.path.realpath(os.path.join(git_root, Path(path)))),
            changed + new,
        )
    )


def _get_git_root_for_path(path: Path) -> Path:
    if os.path.isdir(path):
        dir_path = path
    else:
        dir_path = os.path.dirname(path)
    try:
        relative_path = (
            subprocess.check_output(
                [
                    "git",
                    "rev-parse",
                    "--show-prefix",
                ],
                cwd=os.path.realpath(dir_path),
                stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        # --show-toplevel doesn't work in some windows environment with posix paths,
        # like msys2, so we have to use --show-prefix instead
        git_root = os.path.abspath(
            os.path.join(dir_path, "../" * len(Path(relative_path).parts))
        )
        # call realpath to resolve symlinks, so all paths match
        return Path(os.path.realpath(git_root))
    except subprocess.CalledProcessError:
        logging.error(f"File {path} isn't part of a git project.")
        raise UserError()


def get_shared_git_root_for_paths(paths: list[Path]) -> Path:
    git_roots = set[Path]()
    for path in paths:
        git_root = _get_git_root_for_path(path)
        git_roots.add(git_root)
    if not paths:
        git_root = _get_git_root_for_path(Path(os.getcwd()))
        git_roots.add(git_root)

    if len(git_roots) > 1:
        logging.error(
            "All paths must be part of the same git project! Projects provided:"
            f" {git_roots}"
        )
        raise UserError()
    elif len(git_roots) == 0:
        logging.error("No git projects provided.")
        raise UserError()

    return git_roots.pop()


def commit(message: str) -> None:
    """
    Commit all unstaged and staged changes
    """
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-m", message])


def get_diff_for_file(target: str, path: Path) -> str:
    """Return commit data & diff for target versus active code"""
    git_root = GIT_ROOT.get()

    try:
        diff_content = subprocess.check_output(
            ["git", "diff", "-U0", f"{target}", "--", path], cwd=git_root, text=True
        ).strip()
        return diff_content
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining diff for commit '{target}'.")
        raise UserError()


def get_treeish_metadata(target: str) -> dict[str, str]:
    git_root = GIT_ROOT.get()

    try:
        commit_info = subprocess.check_output(
            ["git", "log", target, "-n", "1", "--pretty=format:%H %s"],
            cwd=git_root,
            text=True,
        ).strip()

        # Split the returned string into the hash and summary
        commit_hash, commit_summary = commit_info.split(" ", 1)
        return {"hexsha": commit_hash, "summary": commit_summary}
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining commit data for target '{target}'.")
        raise UserError()


def get_files_in_diff(target: str) -> list[Path]:
    """Return commit data & diff for target versus active code"""
    git_root = GIT_ROOT.get()

    try:
        diff_content = subprocess.check_output(
            ["git", "diff", "--name-only", f"{target}", "--"], cwd=git_root, text=True
        ).strip()
        if diff_content:
            return [Path(path) for path in diff_content.split("\n")]
        else:
            return []
    except subprocess.CalledProcessError:
        logging.error(f"Error obtaining diff for commit '{target}'.")
        raise UserError()


def check_head_exists() -> bool:
    git_root = GIT_ROOT.get()

    try:
        subprocess.check_output(["git", "rev-parse", "HEAD", "--"], cwd=git_root)
        return True
    except subprocess.CalledProcessError:
        return False


def get_default_branch() -> str:
    git_root = GIT_ROOT.get()

    try:
        # Fetch the symbolic ref of HEAD which points to the default branch
        default_branch = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=git_root, text=True
        ).strip()
        return default_branch
    except subprocess.CalledProcessError:
        # Handle error if needed or raise an exception
        raise Exception("Unable to determine the default branch.")



import math
from pathlib import Path


class Interval:
    def __init__(
        self,
        start: int | float,
        end: int | float,
    ):
        self.start = start
        self.end = end

    def contains(self, line_number: int):
        return self.start <= line_number <= self.end


def parse_intervals(interval_string: str) -> list[Interval]:
    try:
        intervals = list[Interval]()
        for interval in interval_string.split(","):
            interval = interval.split("-", 1)
            if len(interval) == 1:
                intervals += [Interval(int(interval[0]), int(interval[0]))]
            else:
                intervals += [Interval(int(interval[0]), int(interval[1]))]
        return intervals
    except (ValueError, IndexError):
        return []


class CodeFile:
    """
    Represents a file along with the lines that should be in the prompt context. Can be
    iniitialized with a Path object or a string path with optional line information
    such as "path/to/file.py:1-5,7,12-40".

    Attributes:
        path: The absolute path to the file.
        intervals: The lines in context.
    """

    def __init__(self, path: str | Path):
        if Path(path).exists():
            self.path = Path(path)
            self.intervals = [Interval(0, math.inf)]
        else:
            path = str(path)
            split = path.rsplit(":", 1)
            self.path = Path(split[0])
            if not self.path.exists():
                self.path = Path(path)
                self.intervals = [Interval(0, math.inf)]
            else:
                self.intervals = parse_intervals(split[1])

    def contains_line(self, line_number: int):
        return any([interval.contains(line_number) for interval in self.intervals])



import asyncio
from collections import deque

from termcolor import colored

from mentat.session_stream import SESSION_STREAM


class StreamingPrinter:
    def __init__(self):
        self.strings_to_print = deque[str]([])
        self.chars_remaining = 0
        self.finishing = False
        self.shutdown = False

    def add_string(self, string: str, end: str = "\n", color: str | None = None):
        if self.finishing:
            return

        if len(string) == 0:
            return
        string += end

        colored_string = colored(string, color) if color is not None else string

        index = colored_string.index(string)
        characters = list(string)
        characters[0] = colored_string[:index] + characters[0]
        characters[-1] = characters[-1] + colored_string[index + len(string) :]

        self.strings_to_print.extend(characters)
        self.chars_remaining += len(characters)

    def sleep_time(self) -> float:
        max_finish_time = 1.0
        required_sleep_time = max_finish_time / (self.chars_remaining + 1)
        max_sleep = 0.002 if self.finishing else 0.006
        min_sleep = 0.002
        return max(min(max_sleep, required_sleep_time), min_sleep)

    async def print_lines(self):
        stream = SESSION_STREAM.get()

        while not self.shutdown:
            if self.strings_to_print:
                next_string = self.strings_to_print.popleft()
                await stream.send(next_string, end="", flush=True)
                self.chars_remaining -= 1
            elif self.finishing:
                break
            await asyncio.sleep(self.sleep_time())

    def wrap_it_up(self):
        self.finishing = True

    def shutdown_printer(self):
        self.shutdown = True



import json
import os
from collections import defaultdict

import fire
import pytest
from termcolor import cprint

benchmark_location = "tests/benchmark_test.py"
benchmark_log_location = "benchmark_results/benchmark.log"
benchmark_results_location = "benchmark_results/benchmark.results"


def main(run=False, threshold=0.75, count=1):
    os.makedirs("benchmark_results", exist_ok=True)

    if not os.path.isfile(benchmark_results_location):
        with open(benchmark_results_location, "w") as results_file:
            results = {}
            results_json = json.dumps(results)
            results_file.write(results_json)

    with open(benchmark_results_location, "r") as results_file:
        results = json.load(results_file)

    if run:
        for _ in range(count):
            os.environ["MENTAT_BENCHMARKS_RUNNING"] = "true"
            pytest.main(
                [
                    benchmark_location,
                    "--benchmark",
                    "--report-log",
                    benchmark_log_location,
                ]
            )
            os.environ["MENTAT_BENCHMARKS_RUNNING"] = "false"
            print()
            nodes = []
            with open(benchmark_log_location, "r") as result_file:
                for line in result_file:
                    nodes.append(json.loads(line))

            # pytest_reportlog makes multiple nodes per test
            test_names = defaultdict(lambda: True)
            for node in nodes:
                if "location" in node:
                    test_name = node["location"][2]
                    passed = node["outcome"] == "passed"
                    test_names[test_name] = test_names[test_name] and passed

            for test_name, success in test_names.items():
                if test_name not in results:
                    results[test_name] = [0, 0]
                results[test_name][1] += 1
                if success:
                    results[test_name][0] += 1
                    cprint(f"{test_name} passed", "green")
                else:
                    cprint(f"{test_name} failed", "red")
            print()

            with open(benchmark_results_location, "w") as results_file:
                results_json = json.dumps(results)
                results_file.write(results_json)

    for test_name in results.keys():
        passes, total = results[test_name]
        frac = passes / total
        cprint(
            f"{test_name}: {passes} / {total} {frac * 100}%",
            "red" if frac < threshold else "green",
        )


if __name__ == "__main__":
    fire.Fire(main)
